//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//
// "Creature Box" -- flocking app
//
// by Christopher Rasmussen
//
// CISC 440/640, March, 2014
//
// version 1.0
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#include "Utils.hh"
#include "Flocker.hh"
#include "Predator.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// to avoid gimbal lock issues...

#define MAX_LATITUDE_DEGS     89.0
#define MIN_LATITUDE_DEGS    -89.0

#define BOX_WIDTH  9.0
#define BOX_HEIGHT 5.0
#define BOX_DEPTH  7.0

#define CAMERA_MODE_ORBIT    1
#define CAMERA_MODE_CREATURE 2

#define MIN_ORBIT_CAM_RADIUS    (BOX_DEPTH)
#define MAX_ORBIT_CAM_RADIUS    (25.0)

#define DEFAULT_ORBIT_CAM_RADIUS            9.0
#define DEFAULT_ORBIT_CAM_LATITUDE_DEGS     0.0
#define DEFAULT_ORBIT_CAM_LONGITUDE_DEGS    90.0

#define MIN_UPDATE_PERIOD_MS                1.0

//----------------------------------------------------------------------------

double orbit_cam_radius = DEFAULT_ORBIT_CAM_RADIUS;
double orbit_cam_delta_radius = 0.1;

double orbit_cam_latitude_degs = DEFAULT_ORBIT_CAM_LATITUDE_DEGS;
double orbit_cam_longitude_degs = DEFAULT_ORBIT_CAM_LONGITUDE_DEGS;
double orbit_cam_delta_theta_degs = 1.0;

// how long between updates of simulation in milliseconds

int global_num_updates = 0;
double global_update_period_ms = 16.667;  // 60 Hz; period ms WAS 50.0
double global_last_t;

int win_scale_factor = 60;
int win_w = win_scale_factor * 16.0;
int win_h = win_scale_factor * 9.0;
 
int camera_mode = CAMERA_MODE_ORBIT;

int num_flockers = 50;   // 400 is "comfortable" max on my machine

extern int flocker_history_length;
extern int flocker_draw_mode;
extern vector <Flocker *> flocker_array;    
extern vector <vector <double> > flocker_squared_distance;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// has a key been pressed?

void keyboard(unsigned char key, int x, int y)
{
  if (key == 'q')
    exit(1);

  // orbit rotate

  else if (key == 'a') 
    orbit_cam_longitude_degs -= orbit_cam_delta_theta_degs;
  else if (key == 'd')
    orbit_cam_longitude_degs += orbit_cam_delta_theta_degs;
  else if (key == 'w') {
    if (orbit_cam_latitude_degs + orbit_cam_delta_theta_degs <= MAX_LATITUDE_DEGS)
      orbit_cam_latitude_degs += orbit_cam_delta_theta_degs;
  }
  else if (key == 's') {
    if (orbit_cam_latitude_degs - orbit_cam_delta_theta_degs >= MIN_LATITUDE_DEGS)
      orbit_cam_latitude_degs -= orbit_cam_delta_theta_degs;
  }

  // orbit zoom in/out

  else if (key == 'z') {
    if (orbit_cam_radius + orbit_cam_delta_radius <= MAX_ORBIT_CAM_RADIUS)
      orbit_cam_radius += orbit_cam_delta_radius;
  }
  else if (key == 'c') {
    if (orbit_cam_radius - orbit_cam_delta_radius >= MIN_ORBIT_CAM_RADIUS)
      orbit_cam_radius -= orbit_cam_delta_radius;
  }

  // orbit pose reset

  else if (key == 'x') {
    orbit_cam_radius = DEFAULT_ORBIT_CAM_RADIUS;
    orbit_cam_latitude_degs = DEFAULT_ORBIT_CAM_LATITUDE_DEGS;
    orbit_cam_longitude_degs = DEFAULT_ORBIT_CAM_LONGITUDE_DEGS;
  }

  // flocker drawing options

  else if (key == '8')
    flocker_draw_mode = DRAW_MODE_POLY;
  else if (key == '9')
    flocker_draw_mode = DRAW_MODE_AXES;
  else if (key == '0')
    flocker_draw_mode = DRAW_MODE_HISTORY;

  // redraw

  glutPostRedisplay();
}

//----------------------------------------------------------------------------

void global_update(int value)
{
  int i;
  double t, diff_t_ms;

  // precalculate inter-flocker distances

  calculate_flocker_squared_distances();

  // get new_position, new_velocity for each flocker

  for (i = 0; i < flocker_array.size(); i++) 
    flocker_array[i]->update();

  // handle wrapping and make new position, velocity into current

  for (i = 0; i < flocker_array.size(); i++) 
    flocker_array[i]->finalize_update(BOX_WIDTH, BOX_HEIGHT, BOX_DEPTH);

  // post draw event

  glutPostRedisplay();

  // see how we're doing on time

  t = get_timestamp();

  // how long has it been since we were last here?

  diff_t_ms = 1000.0 * (t - global_last_t);
  global_last_t = t;
  global_num_updates++;

  // ZERO excess_time_ms means we are hitting our target of being here 
  // every global_update_period_ms milliseconds

  double excess_time_ms = diff_t_ms - global_update_period_ms;

  // we are running at the desired speed
  
  if (excess_time_ms <= 0.0)
    glutTimerFunc(global_update_period_ms, global_update, 0);

  // we are running too slow...

  else {

    // we can't make up for it, so just run as fast as possible

    if (excess_time_ms > (global_update_period_ms - MIN_UPDATE_PERIOD_MS))
      glutTimerFunc(MIN_UPDATE_PERIOD_MS, global_update, 0);

    // we can make up for it by decreasing the update period

    else 
      glutTimerFunc(global_update_period_ms - excess_time_ms, global_update, 0);

  }

}

//----------------------------------------------------------------------------

// corner "origin" is at (0, 0, 0) -- must translate to center

void draw_box()
{
  glBegin(GL_LINES);

  // "axis" edges are colored

  glColor3f(1, 0, 0);
  glVertex3f(0, 0, 0);
  glVertex3f(BOX_WIDTH, 0, 0);

  glColor3f(0, 1, 0);
  glVertex3f(0, 0, 0);
  glVertex3f(0, BOX_HEIGHT, 0);

  glColor3f(0, 0, 1);
  glVertex3f(0, 0, 0);
  glVertex3f(0, 0, BOX_DEPTH);

  // every other edge is gray

  glColor3f(0.5, 0.5, 0.5);

  glVertex3f(BOX_WIDTH, BOX_HEIGHT, BOX_DEPTH);
  glVertex3f(0, BOX_HEIGHT, BOX_DEPTH);

  glVertex3f(BOX_WIDTH, BOX_HEIGHT, 0);
  glVertex3f(0, BOX_HEIGHT, 0);

  glVertex3f(BOX_WIDTH, 0, BOX_DEPTH);
  glVertex3f(0, 0, BOX_DEPTH);


  glVertex3f(BOX_WIDTH, BOX_HEIGHT, BOX_DEPTH);
  glVertex3f(BOX_WIDTH, 0, BOX_DEPTH);

  glVertex3f(0, BOX_HEIGHT, BOX_DEPTH);
  glVertex3f(0, 0, BOX_DEPTH);

  glVertex3f(BOX_WIDTH, BOX_HEIGHT, 0);
  glVertex3f(BOX_WIDTH, 0, 0);


  glVertex3f(BOX_WIDTH, BOX_HEIGHT, BOX_DEPTH);
  glVertex3f(BOX_WIDTH, BOX_HEIGHT, 0);

  glVertex3f(0, BOX_HEIGHT, BOX_DEPTH);
  glVertex3f(0, BOX_HEIGHT, 0);

  glVertex3f(BOX_WIDTH, 0, BOX_DEPTH);
  glVertex3f(BOX_WIDTH, 0, 0);

  glEnd();
}

//----------------------------------------------------------------------------

void display() 
{
  int i;

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // set up camera

  glPushMatrix();

  if (camera_mode == CAMERA_MODE_ORBIT) {

    double orbit_cam_azimuth = DEG2RAD(orbit_cam_longitude_degs);
    double orbit_cam_inclination = DEG2RAD(90.0 - orbit_cam_latitude_degs);
    
    double x_cam = orbit_cam_radius * sin(orbit_cam_inclination) * cos(orbit_cam_azimuth); // 0.5 * BOX_WIDTH;
    double z_cam = orbit_cam_radius * sin(orbit_cam_inclination) * sin(orbit_cam_azimuth); // 0.5 * BOX_HEIGHT;
    double y_cam = orbit_cam_radius * cos(orbit_cam_inclination); // 15.0;
    
    gluLookAt(x_cam, y_cam, z_cam,
	      0, 0, 0,
	      0, 1, 0);
  }
  else {

    printf("only orbit camera mode currently supported\n");
    exit(1);

  }

  // render scene

  glPushMatrix();

  // this moves the box with its corner at (0, 0, 0) so that its center is at (0, 0, 0)

  glTranslatef(-0.5 * BOX_WIDTH, -0.5 * BOX_HEIGHT, -0.5 * BOX_DEPTH);


  // box

  draw_box();

  // flockers

  for (i = 0; i < flocker_array.size(); i++) {
    //    printf("%i: %.3lf %.3lf %.3lf\n", i, flocker_array[i]->x, flocker_array[i]->y, flocker_array[i]->z); fflush(stdout);
    flocker_array[i]->draw();
  }

  glPopMatrix();    // box and flockers
 
  glPopMatrix();    // overall

  // finish up
  
  glutSwapBuffers();

}

//----------------------------------------------------------------------------

void init() 
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluPerspective(50, (float) win_w / (float) win_h, 0.1, 35);

  glEnable(GL_DEPTH_TEST);

  //  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

//----------------------------------------------------------------------------

int main(int argc, char** argv)
{
  // read command-line flags

  for (int i = 1; i < argc; i++) {
    if (!strcmp(argv[i], "-n") || !strcmp(argv[i], "-N"))
      num_flockers = atof(argv[i+1]);
    else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "-H"))
      flocker_history_length = atoi(argv[i+1]);
  }

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  glutInitWindowSize(win_w, win_h);
  glutCreateWindow("creatures");
  init(); 			     

  // simulation objects

  initialize_random();

  flocker_squared_distance.resize(num_flockers);
  flocker_array.clear();

  for (int i = 0; i < num_flockers; i++) {
    flocker_array.push_back(new Flocker(i, 
					  uniform_random(0, BOX_WIDTH), uniform_random(0, BOX_HEIGHT), uniform_random(0, BOX_DEPTH),
					  uniform_random(-0.01, 0.01), uniform_random(-0.01, 0.01), uniform_random(-0.01, 0.01),
					  0.002,            // randomness
					  0.05, 0.5, 0.02,  // min, max separation distance, weight
					  0.5,  1.0, 0.001, // min, max alignment distance, weight
					  1.0,  1.5, 0.001, // min, max cohesion distance, weight
					  1.0,  1.0, 1.0,
					  flocker_history_length));

    flocker_squared_distance[i].resize(num_flockers);
  }

  // register callback routines

  glutDisplayFunc(display);
  glutKeyboardFunc(keyboard);

  global_last_t = get_timestamp();
  glutTimerFunc(global_update_period_ms, global_update, 0);

  glutMainLoop(); 		    // enter event-driven loop

  return 1;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
